Unit Mouse;                  { Simple mouse unit by Vulture/OT }

INTERFACE

Type  CursorData = Array [1..32] of Word;

Const ArrowMask: Cursordata =
             ($7fff,$3fff,$1fff,$0fff,
              $07ff,$03ff,$01ff,$00ff,
              $007f,$03ff,$03ff,$29ff,
              $71ff,$f0ff,$faff,$f8ff,
              $8000,$C000,$A000,$9000,
              $8800,$8400,$8200,$8100,
              $8f80,$9400,$b400,$d200,
              $8a00,$0900,$0500,$0700);

      HourGlassMask: CursorData =
             ($0000,$0000,$0000,$c003,
              $e007,$f00f,$F81F,$fc3f,
              $fc3F,$F81F,$F00F,$e007,
              $c003,$0000,$0000,$0000,
              $0000,$7ffe,$0000,$1ff8,
              $0ff0,$0000,$0000,$0000,
              $0180,$0340,$07e0,$0e78,
              $1818,$0000,$7ffe,$0000);

Var Mouses: Record           { Format of mouse }
              X,Y: Word;
              Buttons: Word;
            End;

Procedure InitMouse;
Procedure ShowMouse;
Procedure HideMouse;
Procedure MouseWhere;
Procedure SetMouseCursor(X,Y: Word);
Procedure MouseButtons;
Procedure LimitMouse(TopX, TopY, BotX, BotY: Word);
Procedure MouseInfo(VAR X,Y: Integer; VAR Lb,Rb: Boolean);
Procedure SetMouseMask(CursorMask: CursorData);

IMPLEMENTATION

Uses Dos;

{ Initialize the mouse driver }
Procedure InitMouse; Assembler;
Asm
    xor ax,ax                { Function 0 }
    int 33h
End;

{ Show the mouse cursor on the screen }
Procedure ShowMouse; Assembler;
Asm
    mov ax,1                 { Function 1 }
    int 33h
End;

{ Remove the mouse cursor from the screen }
Procedure HideMouse; Assembler;
Asm
    mov ax,2                 { Function 2 }
    int 33h
End;

{ Find the location of the mouse }
Procedure MouseWhere; Assembler;
Asm
    mov ax,3                 { Function 3 }
    int 33h
    mov mouses.x,cx
    mov mouses.y,dx
End;

{ Set new x,y position for mouse }
Procedure SetMouseCursor(X,Y: Word); Assembler;
Asm
    mov ax,4
    mov cx,X
    mov dx,Y
    int 33h
End;

{ Return the press state of the mouse buttons }
Procedure MouseButtons; Assembler;
Asm
    mov ax,5                 { Function 5 }
    xor bx,bx                { All buttons }
    int 33h
    mov mouses.buttons,ax
End;

{ Limit the area the mouse can move to }
Procedure LimitMouse(TopX, TopY, BotX, BotY: Word); Assembler;
Asm
    mov ax,7                 { Function 7 - set X limits }
    mov cx,topx
    mov dx,botx
    int 33h
    mov ax,8                 { Function 8 - set Y limits }
    mov cx,topy
    mov dx,boty
    int 33h
End;

Procedure MouseInfo(VAR X,Y: Integer; VAR Lb,Rb: Boolean); Assembler;
Asm
    mov	ax,3         { Function 3 - Find location }
    int	33h
    les	si,x
    mov	[es:si],cx   { Store x }
    les	si,y
    mov	[es:si],dx   { Store y }

    mov	ax,bx
    and	al,1
    les	si,lb
    mov	[es:si],al   { Store state of left button }
    shr	bl,1
    and	bl,1
    les	si,rb
    mov	[es:si],bl   { Store state of right button }
End;

Procedure SetMouseMask(CursorMask: CursorData);
Var Regs: Registers;
Begin
  Regs.AX := $0009;
  Regs.BX := $0004;
  Regs.CX := $0004;
  Regs.ES := Seg(CursorMask);
  Regs.DX := Ofs(CursorMask);
  Intr($33,Regs);
End;

END.                         { Mouse unit }


