Unit Pcx; { Decodes low and high resolution pcxfiles. By Vulture/OT (c) 1997 }

INTERFACE

Procedure Decode_Pcx(S:String; File_Offset,File_Size: Longint; Where:Word);
Procedure DecodePcx_16(S: String; VgaPos: Longint);
Procedure DecodePcx_Vesa(S:String);

IMPLEMENTATION

{  }

Procedure Set_Color(Color,R,G,B: Byte); Assembler;
Asm
    mov     dx,03c8h
    mov     al,Color            { Load colorvalue (0..255) to alter }
    out     dx,al
    inc     dx
    mov     al,R                { Red }
    out     dx,al
    mov     al,G                { Green }
    out     dx,al
    mov     al,B                { Blue }
    out     dx,al
End;

Procedure Decode_Pcx(S:String; File_Offset,File_Size: Longint; Where:Word);
Var Loop1: Word;                        { Loopcounter variable }
    F: File;                            { File }
    Temp: Pointer;                      { Memory pointer for data }
    RGB_Values: Array[0..767] of Byte;  { Array for all RGB values }
Begin
  Assign(F,S);                          { Name of file }
  Reset(F,1);                           { Open the file }

  If File_Size=0 then File_Size := Filesize(F);

  Seek(F,(File_Offset+File_Size)-768);  { Palette data starts here }
  BlockRead(F,RGB_Values,768);          { Read all RGB values at once }
  For Loop1 := 0 to 255 Do              { Set all colors }
    Set_Color(Loop1,RGB_Values[Loop1*3] shr 2,RGB_Values[Loop1*3+1] shr 2,RGB_Values[Loop1*3+2] shr 2);

  Seek(F,File_Offset+128);              { Skip header }

  GetMem(Temp,File_Size-128-768);       { Get memory for file }
  BlockRead(F,Temp^,File_Size-128-768); { Load entire pcxfile }

  Asm
    push  ds             { Save ds on stack }
    mov   ax,Where
    mov   es,ax          { es:di => destination }
    xor   di,di
    xor   ch,ch
    lds   si,Temp        { ds:si => source }
@Decode:
    lodsb                { Load a byte }
    mov   bl,al          { Save into bl }
    and   bl,11000000b   { Check for compression }
    cmp   bl,11000000b
    jne   @SingleWrite

    mov   cl,al
    and   cl,00111111b   { Write counter in cx }
    lodsb                { Load the second byte }
    rep   stosb          { And write the 2nd byte # of times }
    jmp   @EndWrite

@SingleWrite:
    stosb                { Write uncompressed byte }
@EndWrite:
    cmp   di,64000       { Segment limit }
    jb    @Decode

    pop   ds             { Restore ds }
  End;
  FreeMem(Temp,File_Size-128-768);      { Free used memory }
  Close(F);                             { Close the file }
End;

{  }

Const Buffer = 65528;       { Largest possible }

Var   Temp: Pointer;        { Memory pointer for data }
      HoriCount, Vga_Index, DataCount: Word;
      Mode,Bytes_Per_Line,Plane,Counter: Byte;
      Header: Array[0..127] of Byte; { Array for header data }

Procedure Decode_It; Assembler;
Asm
   push    ds                   { Save stuff on stack }
   push    bp

{ === Init stuff for run through buffer === }
   mov     bl,Plane             { First plane }
   mov     bh,Bytes_Per_Line    { # of bytes on 1 scanline }
   mov     bp,DataCount         { # of bytes to process }
   xor     cx,cx
   mov     cl,Counter           { Single/multiple write }
   mov     dx,HoriCount         { Scanline pointer }
   mov     di,Vga_Index         { Where to start }
   cld                          { Clear direction flag }

{ === Set ds & es === }
   mov     ax,0a000h
   mov     es,ax                { Destination => es:di }
   lds     si,Temp              { Source => ds:si (ds=destoyed!) }

{ === Check on multiple write === }
   cmp     cl,0
   je      @Get_Data            { If cl = 0 then process next byte }
   lodsb                        { Else cl = counter so continue loop }
   jmp     @Write_Byte

{ === Start of decoding === }
@Get_Data:
   cmp     si,bp                { Check if more data is to be decoded }
   je      @Quit_Decode         { Quit on end of buffer or no more data }
   lodsb                        { Get byte from ds:si }

   cmp     al,11000000b         { Is it a loop byte? }
   jb      @Single_Write        { If not, write a single byte }
   mov     cl,al                { Else setup for multiple write }
   and     cl,00111111b         { Set loopcounter for multiple bytes }

   cmp     si,bp                { Range check }
   je      @Quit_Decode
   lodsb                        { Load next byte }

   jmp     @Write_Byte

@Single_Write:
   mov     cl,1                 { Set loopcounter for 1 byte }

@Write_Byte:                    { Write byte(s) to vga }
   stosb
   inc     dl                   { Increment collumn }
   cmp     dl,bh                { End of line? }
   je      @Reached_End         { If so, set other plane }
   loop    @Write_Byte          { If not, write another byte }
   jmp     @Get_Data            { If all bytes written, fetch new data }

@Reached_End:
   shl     bl,1                 { Next plane }
   cmp     bl,00001000b         { Is this the last plane? }
   jle     @Not_Last            { If not, reset scanline }
   shr     bl,4                 { If so, reset plane & next scanline }
   jmp     @Set_Plane

@Not_Last:
   sub     di,dx                { Reset to start of scanline }

@Set_Plane:
   push    ax                   { Save byte }
   cli
   mov     dx,03c4h
   mov     al,02h               { Map Mask Register }
   mov     ah,bl                { Load plane in ah }
   out     dx,ax
   sti
   pop     ax                   { Restore byte }
   xor     dx,dx                { Reset collumn counter }
   loop    @Write_Byte          { And write next byte or get data }
   jmp     @Get_Data

@Quit_Decode:

   pop     bp                   { Restore from stack }
   pop     ds

{ === Save variables === }
   mov     Plane,bl
   mov     Vga_Index,di
   mov     Counter,cl
   mov     HoriCount,dx
End;

Procedure Set_Palette; Assembler;
Asm
   push    ds

{ === Set ds:si === }
   lea     si,Header
   add     si,16                { Pal data starts at offset 16 in header }

{ === Start at color 0 === }
   mov     dx,03c8h
   xor     al,al
   out     dx,al
   inc     dx

{ === Send 48 bytes === }
   mov     cx,48
@PalLoop:
   mov     al,byte ptr ds:[si]
   shr     al,2
   out     dx,al
   inc     si
   loop    @PalLoop

{ === Init palette registers === }
   xor     dl,dl
   mov     cx,16
@PalRegLoop:
   mov     bl,dl                { Register }
   mov     bh,dl                { Value }
   mov     ax,1000h
   int     10h                  { Let bios handle this }
   inc     dl
   loop    @PalRegLoop

   pop     ds
End;

Procedure DecodePcx_16(S: String; VgaPos: Longint);
Var F: File;                            { File }
Begin

{ === Open file and set recordsize 1 === }
  {$I-}
  Assign(F,S);
  Reset(F,1);

{ === Read header into array === }
  BlockRead(F, Header, SizeOf(Header));
  Set_Palette;                          { Init colors }

{ === Set various variables === }
  Plane := 1;                           { First plane first... }
  Bytes_Per_Line := Header[66];         { Read # of bytes per scanline }
  Vga_Index := VgaPos;
  Counter := 0;

{ === Write to vga === }
  Getmem(Temp, Buffer);                 { Allocate memory }

  Repeat                                { Breaking the 64kB limit }
    BlockRead(F, Temp^, Buffer, DataCount);
    Decode_It;
  Until Eof(F);

{ === Free memory and close file === }
  Freemem(Temp, Buffer);                { Free allocated memory }
  Close(F);
  {$I+}
End;

{  }

Type TVesaMode =
     Record
       Attr     : Word;         { Mode Attributes                   }
       WinA     : Byte;         { Window A attributes               }
       WinB     : Byte;         { Window B attributes               }
       Gran     : Word;         { Window granularity in K bytes     }
       WinSiz   : Word;         { Size of window in K bytes         }
       SegA     : Word;         { Segment address of window A       }
       SegB     : Word;         { Segment address of window B       }
       WinFunc  : Procedure;    { Windows positioning function      }
       Bytes    : Word;         { Number of bytes per line          }
       Width    : Word;         { Number of horizontal pixels       }
       Height   : Word;         { Number of vertical pixels         }
       CharW    : Byte;         { Width of character cell           }
       CharH    : Byte;         { Height of character cell          }
       Planes   : Byte;         { Number of memory planes           }
       Bits     : Byte;         { Number of bits per pixel          }
       nBanks   : Byte;         { Number of banks        (not used) }
       Model    : Byte;         { Memory model type                 }
       Banks    : Byte;         { Size of bank           (not used) }
       Pages    : Byte;         { Number of image pages             }
       Reserved : Byte;         { The following are for 15,16,24,32 bit colour modes }
       RedMaskSize   : Byte;    { Size of Red mask in bits          }
       RedFieldPos   : Byte;    { Bit position of LSB of Red mask   }
       GreenMaskSize : Byte;    { Size of Green mask in bits        }
       GreenFieldPos : Byte;    { Bit position of LSB of Green mask }
       BlueMaskSize  : Byte;    { Size of Blue mask in bits         }
       BlueFieldPos  : Byte;    { Bit position of LSB of Blue mask  }
       RsvdMaskSize  : Byte;    { Size of Reserved mask in bits     }
       RsvdFieldPos  : Byte;    { Bit pos. of LSB of Reserved mask  }
       DirColModeInf : Byte;    { Direct Colour mode attributes     }
       Filler   : Array[0..215] of Byte; { Not used - filler        }
     End;

Type Buffer_Mem = Array[1..65528] of Byte;          { Buffer memory }

Var  VesaMode : TVesaMode;
     Actbank  : Word;           { Number of actual bank on the screen }

Procedure Init_Vesa_Mode; Assembler;
Asm
    mov  ax,4F02h
    mov  bx,101h         {mode to set is 101h for 640x480x256 (112h for 16m color)}
    int  10h             {set it}
    cmp  ax,4Fh          {if setting wasn't done}
    jne  @False_Mode     {then jump to modefalse}

    mov  ax,4F01h        {else load info about this vesa mode into vesamode}
    mov  cx,101h
    lea  di,Vesamode
    int  10h

@False_Mode:
End;

Procedure Set_Vesa_Bank; Assembler;
Asm
    pusha
    mov     ax,Actbank
    shl     ax,6
    xor     dx,dx
    div     vesamode.gran
    mov     cx,ax
    mov     dx,ax
    mov     ax,4F05h
    xor     bx,bx
    int     10h
    mov     ax,4F05h
    mov     bx,1
    int     10h
    popa
    ret
end;

Procedure Put_Vesa_Pixel(x,y:Integer; Color:Byte); Assembler;
Asm
    mov     ax,sega000
    mov     es,ax
    xor     dx,dx
    mov     ax,word ptr [y]
    mov     bx,640
    mul     bx                  { ax:=ax*bx (here, dx gets the num. of the needed bank) }
    mov     di,ax               { di:=ax }
    add     di,word ptr [x]     { di:=di+x }
    adc     dx,0                { Add the carry flag to dx }
    cmp     dx,actbank          { Is the needed bank the actual? }
    jz      @Put_it
    mov     Actbank,dx          { Make the needed the actual }
    call    Set_Vesa_Bank
@Put_it:
    mov     al,byte ptr [Color]
    mov     byte ptr es:[di],al
End;

Procedure Clear_Vesa_Screen(Color: Byte); Assembler;
Asm
    mov     ax,sega000
    mov     es,ax
    xor     di,di
    mov     al,Color           { Load color into ax }
    mov     ah,al
    xor     dx,dx
    mov     bx,5               { There are 5 banks to fill }
@Bank_Loop:
    mov     Actbank,dx         { Save actual bank number }
    call    Set_Vesa_Bank
    mov     cx,32768           { # of words per bank }
@Fill_Loop:
    mov     word ptr es:[di],ax
    inc     di
    inc     di
    loop    @Fill_Loop         { Fill complete bank }
    inc     dx
    dec     bx
    jnz     @Bank_Loop         { Until all banks are filled }
End;

Function Get_Vesa_Hires(x,y:Integer): Byte; Assembler;
Asm
    mov     ax,sega000
    mov     es,ax
    xor     dx,dx
    mov     ax,word ptr [y]
    mov     bx,640
    mul     bx
    mov     di,ax
    add     di,word ptr [x]
    adc     dx,0
    cmp     dx,Actbank
    jz      @Put_it
    mov     Actbank,dx
    call    Set_Vesa_Bank
@Put_it:
    mov     al,byte ptr es:[di]
End;

Procedure DecodePcx_Vesa(S:String);           { Decodes 640x480x256 .pcx files }
Var F: File;
    T: ^Buffer_Mem;
    Loop1, Loop2: Word;
    Temp1, Temp2: Byte;
    X,Y: Word;
    Colors: Array[0..767] of Byte;            { Array for all RGB values }
    Num_Read,Count: Word;
Begin
  Assign(F,S);                                { Open file and set record size }
  Reset(F,1);

  For Loop1 := 0 to 255 Do                    { Reset all colors to black }
    Set_Color(Loop1,0,0,0);

  X := 0;
  Y := 0;
  Seek(F,128);                                { Skip header }

  Getmem(T,Sizeof(Buffer_Mem));               { Resever memory for Buffer_Mem }

  Repeat
    Blockread(F,T^,Sizeof(Buffer_Mem),Num_Read);           { Read a Buffer_Mem full of data }
    If Num_Read<Sizeof(Buffer_Mem) then Dec(Num_Read,768); { Don't decode palette data }

    For Loop1 := 1 to Num_Read do             { Decode the entire Buffer_Mem }
    Begin
      Temp1 := T^[Loop1];                     { Get a byte from the Buffer_Mem }
      If Temp1 and 192 = 192 then             { Compressed byte? }
      Begin
        If Loop1<Num_Read then                { Process the last byte }
        Begin
          Inc(Loop1);                         { Increase Buffer_Mem counter }
          Temp2 := T^[Loop1];
          For Loop2 := 1 to (Temp1 and 63) Do { Plot # of bytes }
          Begin
            Put_Vesa_Pixel(X,Y,Temp2);
            X := (X+1) mod 640;               { Increase x position }
            If X = 0 then Inc(Y);             { If x was reset, then do next scanline }
          End;
        End else Seek(F,Filepos(F)-1);        { Goto correct position }
      End else
      Begin                                   { No compression, plot byte }
        Put_Vesa_Pixel(X,Y,Temp1);
        X := (X+1) mod 640;
        If X = 0 then Inc(Y);
      End;
    End;
  Until Eof(F);                               { Until complete file decode }

  Seek(F,FileSize(F)-768);                    { Palette data starts here }
  BlockRead(F,Colors,768);                    { Read all RGB values at once }
  For Loop1 := 0 to 255 Do                    { Set all colors }
    Set_Color(Loop1,Colors[Loop1*3] shr 2,Colors[Loop1*3+1] shr 2,Colors[Loop1*3+2] shr 2);

  Close(F);                                   { Close the file }
  Freemem(T,Sizeof(Buffer_Mem));
End;


End.