
{  }

Program Ray_Casting_Engine;               { By Vulture/OT 1998 }

Uses Crt,Gfx,Pcx,Mouse;                   { Used units }

Type Wall_Type = Record
                   Dist      : Longint;   { Distance from viewer }
                   Texture   : Word;      { Texture segment value }
                   Xoffset   : Word;      { X-offset into texture }
                   MapValue  : Byte;      { The map number }
                   Xmap,Ymap : Byte;      { 2d map positions }
                 End;

     Door_Type = Record
                   Status    : Byte;      { 0=closed, 1=opening, 2=opened }
                   Delays    : Word;
                   Yoffset   : Word;      { Y-offset into texture }
                   Speed     : Byte;
                   Xmap,Ymap : Byte;
                 End;

     Lint_Table = Array[0..1920] of Longint;

Const WallMap: Array[0..31,0..31] of Byte =   { Walls layout in grids }
      {$I walls.pas}

      FloorMap: Array[0..31,0..31] of Byte =  { Floor layout in grids }
      {$I floor.pas}

      LeftX = 6;                          { Viewport clipping constants }
      RiteX = 239;
      TopY = 6;
      BotY = 193;
      ClipHeight = 200-TopY-(200-BotY);

      Angle_0    = 0;                     { Much used angles }
      Angle_30   = 160;
      Angle_90   = 480;
      Angle_180  = 960;
      Angle_270  = 1440;
      Angle_330  = 1760;
      Angle_360  = 1920;

      Grid_Size  = 64;                    { Grid size in units/pixels }
      Horizon    = 100;                   { Middle of viewport }
      Speed      = 11;                    { Forward/backward maximum speed }
      Turning    = 40;                    { Turning speed }
      Door_Speed = 2;                     { Movement speed of doors }
      Close_Wall = 32;                    { How close the player gets to a wall }

      Door_Code  = 12;                    { Starting texture number of doors }
      Trans_Code = 11;                    { Starting texture of transparent walls }

      Max_Text   = 15;                    { Max # of textures }
      Max_Walls  = 10;                    { Max # of transparent walls }
      Max_Doors  = 10;                    { Max # of active doors }

      Texture_Size = 4096;                { Texturesize in bytes }

Var Tangens: Array[0..1920] of Real;      { Tangens table }
    NextX: Array[0..1920] of Real;        { NextX/NextY tables are used for }
    NextY: Array[0..1920] of Real;        { grabbing raycasting step values }
    Cosine: ^Lint_Table;                  { Cosine table }
    Sine: ^Lint_Table;                    { Sine table }

    Floor_Row_Table: Array[1..100] of Longint;

    Px,Py,Pd: Integer;                    { Player x,y,degree }
    Tx,Ty: Integer;                       { Test x,y values }
    Move_Value: Integer;                  { Movement speed }

    XmapX,YmapX: Byte;                    { 2d map coordinates }
    XmapY,YmapY: Byte;                    { 2d map coordinates }
    TextX,TextY: Byte;                    { Texture numbers for xray,yray }
    Xdist,Ydist: Longint;                 { Distance of x,y walls }
    RayXofs,RayYofs: Integer;             { Offset for texture mapping }
    Wall_Found: Boolean;                  { Used to find walls }

    ShadeFactor: Byte;                    { Light shading factor }
    MapNr: Byte;                          { Map value used in main proc }
    Loop1,Loop2,Loop3: Integer;           { All purpose variables }

    TexPoint: Array[1..Max_Text] of Pointer;    { All textures (1 texture=4096 bytes }
    Textures: Array[1..Max_Text] of Word;       { All segment values of textures }
    Palette: Array[1..768] of Byte;             { Palette data }

    TempMap: Array[0..31,0..31] of Byte;        { Temp level layout in grids }
    Walls: Array[1..Max_Walls] of Wall_Type;    { # of transparent walls allowed }
    Doors: Array[1..Max_Doors+1] of Door_Type;  { # of active doors allowed }

    Active_Doors: Byte;                   { # of doors opening or closing }
    Door_Number: Byte;

    Mx,My: Integer;                       { Mouse info }
    Lb,Rb: Boolean;
    Key: Char;                            { Keyboard interceptor }

{$I MISC.PAS}

{  }

Procedure Calculate_Tables;
Var Angle,Temp: Real;
Begin
  New(Cosine);
  New(Sine);

  Angle := 0;
  For Loop1 := 0 to 1920 do               { Calculate Tan/Cos/Sin/NextX/NextY tables }
  Begin
    Temp := (pi/180)*Angle;               { Convert from radians to degrees }
    Angle := Angle + 0.1875;
    If (Loop1 <> Angle_90) and (Loop1 <> Angle_270) and
       (Loop1 <> Angle_0) and (Loop1 <> Angle_180) and (Loop1 <> Angle_360) then
    Begin
      Tangens[Loop1] := Sin(Temp) / Cos(Temp);
      NextX[Loop1] := 64 / Tangens[Loop1];
      NextY[Loop1] := 64 * Tangens[Loop1];
    End;
    Temp := (Loop1*pi)/Angle_180;         { Fixed point values }
    Cosine^[Loop1] := Round(Cos(Temp)*1024);
    Sine^[Loop1] := Round(Sin(Temp)*1024);
  End;

  For Loop1 := 100 downto 1 do
    Floor_Row_Table[Loop1] := Round((7500*1024) / (Loop1));
End;

{  Raycasting procedures  }

{ Find vertical walls (x-axis) }
Function Xray(x,y: Longint; Angle: Integer): Byte;
Var Xpos: Longint;                        { World coordinates }
    Ypos: Real; rYpos: Longint;
    Xstep: Longint;                       { Step values for ray }
    Ystep: Real;
Begin
  If (Angle=Angle_90) or (Angle=Angle_270) then Exit;
  Wall_Found := False;

{ Set start positions of ray }
  If (Angle>Angle_270) or (Angle<Angle_90) then Xpos := (x shr 6) shl 6 + 64
                                           else Xpos := (x shr 6) shl 6 - 1;
  Ypos := y + (Xpos-x) * Tangens[Angle];

{ Set step values of ray }
  If (Angle>Angle_90) and (Angle<Angle_270) then
  Begin
    Xstep := -Grid_Size;
    Ystep := -NextY[Angle];
  End else
  Begin
    Xstep := Grid_Size;
    Ystep := NextY[Angle];
  End;

{ Let's find a wall }
  Repeat
    rYpos := Round(Ypos);
    XmapX := rYpos shr 6;                 { Get map coordinates }
    YmapX := Xpos shr 6;
    If TempMap[XmapX,YmapX] > 0 then      { Found a wall at these coordinates? }
    Begin
      If TempMap[XmapX,YmapX] > 10 then   { Transparent wall; displace! }
      Begin
        Xdist := (Xpos+(Xstep shr 1)-x) shl 10 div Cosine^[Angle];
        RayXofs := Round(rYpos+(Ystep / 2)) mod 64;
      End else
      Begin
        Xdist := (Xpos-x) shl 10 div Cosine^[Angle];  { Calculate distance from player }
        RayXofs := rYpos mod 64;          { Calculate texture offset }
      End;
      Wall_Found := True;
    End;
    Inc(Xpos,Xstep);                      { Update ray position }
    Ypos := Ypos + Ystep;
  Until Wall_Found = True;

{ Return value and exit }
  Xray := TempMap[XmapX,YmapX];           { Return texture number }
End;

{ Find horizontal walls (y-axis) }
Function Yray(x,y: Longint; Angle: Integer): Byte;
Var Xpos: Real; rXpos: Longint;
    Ypos: Longint;
    Xstep: Real;
    Ystep: Longint;
Begin
  If (Angle=Angle_0) or (Angle=Angle_180) then Exit;
  Wall_Found := False;

{ Set start positions of ray }
  If (Angle>Angle_180) then Ypos := (y shr 6) shl 6 - 1
                       else Ypos := (y shr 6) shl 6 + 64;

  If (Angle=Angle_90) or (Angle=Angle_270)               { ?!? }
    then Xpos := x
    else Xpos := x + (Ypos-y) / Tangens[Angle];

{ Set step values of ray }
  If (Angle<Angle_180) then
  Begin
    Ystep := Grid_Size;
    Xstep := NextX[Angle];
  End else
  Begin
    Ystep := -Grid_Size;
    Xstep := -NextX[Angle];
  End;

{ Let's find a wall }
  Repeat
    rXpos := Round(Xpos);
    XmapY := Ypos shr 6;                  { Get map coordinates }
    YmapY := rXpos shr 6;
    If TempMap[XmapY,YmapY] > 0 then      { Found a wall at these coordinates }
    Begin
      If TempMap[XmapY,YmapY] > 10 then   { Transparent wall; displance! }
      Begin
        Ydist := (Ypos+(Ystep shr 1)-y) shl 10 div Sine^[Angle];
        RayYofs := Round(rXpos+(Xstep / 2)) mod 64;
      End else
      Begin
        Ydist := (Ypos-y) shl 10 div Sine^[Angle]; { Calculate distance from player }
        RayYofs := rXpos mod 64;          { Calculate texture offset }
      End;
      Wall_Found := True;
    End;
    Xpos := Xpos + Xstep;                 { Update ray position }
    Inc(Ypos,Ystep);
  Until Wall_Found = True;

{ Return value and exit }
  Yray := TempMap[XmapY,YmapY];           { Return texture number }
End;

{  Door routines  }

Function Find_Door(X,Y: Byte): Byte;      { Check if door is already in list }
Var Door_Found: Boolean;
Begin
  Door_Number := 0;                       { The final number of the active door }
  Door_Found := False;
  Repeat
    Inc(Door_Number);
    If (Doors[Door_Number].Xmap = X) and (Doors[Door_Number].Ymap = Y) then Door_Found := True;
  Until (Door_Number=Active_Doors) or (Door_Found=True);
  If Door_Found then Find_Door := 1 else Find_Door := 0;
End;

Procedure Check_Door(X,Y,D: Longint);     { Check if a door is in front of the player }
Var Find: Byte;
Begin
  Xdist := MaxLongInt;                    { Set distances to max length }
  Ydist := MaxLongInt;

  TextX := Xray(X,Y,D);                   { Cast x,y rays towards walls/doors }
  TextY := Yray(X,Y,D);

  If (TextX<Door_Code) and (TextY<Door_Code) then Exit;   { No door was found }
  If (Xdist>128) and (Ydist>128) then Exit; { Door is too far away }

  If Xdist<Ydist then                     { X door is closer; store it }
  Begin
    Find := Find_Door(XmapX,YmapX);
    If Find=0 then
    Begin
      Inc(Active_Doors);
      Doors[Active_Doors].Status := 1;    { Door is opening now }
      Doors[Active_Doors].Delays := 100;
      Doors[Active_Doors].Yoffset := 0;
      Doors[Active_Doors].Xmap := XmapX;
      Doors[Active_Doors].Ymap := YmapX;
    End;
  End else
  Begin
    Find := Find_Door(XmapY,YmapY);
    If Find=0 then
    Begin
      Inc(Active_Doors);
      Doors[Active_Doors].Status := 1;
      Doors[Active_Doors].Delays := 100;
      Doors[Active_Doors].Yoffset := 0;
      Doors[Active_Doors].Xmap := XmapY;
      Doors[Active_Doors].Ymap := YmapY;
    End;
  End;
End;

Procedure Update_Doors;                   { Update all active doors once per frame }
Var i,j: Byte;
Begin
  For i := 1 to Active_Doors do
  Begin
    Case Doors[i].Status of
      0: Begin                            { Door closed; remove from active list }
           For j := i to Active_Doors do Doors[j] := Doors[j+1];
           If Active_Doors>1 then Dec(Active_Doors);
         End;
      1: If Doors[i].Yoffset<58 then Inc(Doors[i].Yoffset,Door_Speed) else Doors[i].Status:=2;
      2: If Doors[i].Delays>0 then Dec(Doors[i].Delays) else Doors[i].Status := 3;
      3: If Doors[i].Yoffset>0 then Dec(Doors[i].Yoffset,Door_Speed) else Doors[i].Status := 0;
    End;
  End;
End;

{  Graphics routines  }

Procedure Draw_Wall_Slice(ScreenX,ScreenY,Height,Xoffset: Integer; Shade,Yoffset: Byte; Source,Dest: Word);
Var Step,TextureY: Word;
Begin
  Step := 16384 div Height;               { 64 shl 8 div Height }

  If ScreenY < TopY then                  { Calculate y offset into texture }
  Begin
    TextureY := (((TopY-ScreenY) * Step) shr 8)+Yoffset; { Abs((ScreenY*Step) shr 8)+Yoffset }
    ScreenY := TopY;
    Height := ClipHeight;
  End else TextureY := Yoffset;

  Asm
     push  ds

     mov   ax,Dest
     mov   es,ax
     mov   di,ScreenY
     mov   ax,ScreenY
     shl   di,8
     shl   ax,6
     add   di,ax
     add   di,ScreenX

     mov   ax,Source
     mov   ds,ax
     xor   bx,bx                          { Starting y position }

     mov   cx,Height
@LoopY:
     mov   si,bx
     shr   si,8                           { Convert to normal y value (remove fixed point) }
     add   si,TextureY                    { Add y offset into texture }
     cmp   si,63                          { Quit if door is drawn completely }
     ja    @QuitDoor
     shl   si,6                           { Multiply by 64 }
     add   si,Xoffset                     { Add x offset to get final position }

     mov   al,byte ptr ds:[si]            { Get texel }

     cmp   al,0                           { Don't draw if transparent }
     je    @Black
     mov   dl,al
     and   dl,00001111b                   { Color range 0..15 }
     add   dl,Shade                       { Add shade value }
     cmp   dl,15                          { Larger/equal 15? }
     jae   @Black                         { Color is black, don't plot }
     and   al,11110000b                   { Upper range }
     add   al,dl

     mov   byte ptr es:[di],al            { Plot shaded pixel }
@Black:
     add   di,320
     add   bx,Step
     dec   cx
     jnz   @LoopY

@QuitDoor:
     pop   ds
  End;
End;

Procedure Draw_FloorCeiling_Slice(CosValue: Longint; StartY,EndY,StartX,X,Y,Angle,Dest: Word);
Var Distance: Longint;
    Shade: Byte;
    Loop,xv,yv: Integer;
    Texture: Word;
    Ty,By: Word;
Begin
  Ty := (StartY shl 8)+(StartY shl 6)+StartX;        { Y positions on screen }
  By := (EndY shl 8)+(EndY shl 6)+StartX;
  For Loop := EndY to BotY do                        { Y clipping }
  Begin
    Distance := Floor_Row_Table[Loop-100] div CosValue; { EndY is always > 100 }
    Shade := Distance shr ShadeFactor;
    xv := (Distance * Cosine^[Angle] shr 10) + x;       { World coordinates }
    yv := (Distance * Sine^[Angle] shr 10) + y;
    Texture := Textures[FloorMap[yv shr 6,xv shr 6]];   { Get texture segment }

    Asm
      push  ds

      mov   ax,Texture
      mov   ds,ax
      mov   si,yv
      and   si,63
      shl   si,6
      mov   ax,xv
      and   ax,63
      add   si,ax                         { ds:si => texture pixel }

      mov   al,byte ptr ds:[si]           { Get texel }

      mov   dl,al
      and   dl,00001111b                  { Color range 0..15 }
      add   dl,Shade                      { Add shade value }
      cmp   dl,15                         { Larger/equal 15? }
      jae   @Black                        { Color is black, don't plot }
      and   al,11110000b                  { Upper range }
      add   al,dl

      mov   bx,Dest
      mov   es,bx                         { es:di => destination }
      mov   di,By                         { Plot shaded pixels }
      mov   es:[di],al
      mov   di,Ty
      mov   es:[di],al
@Black:

      pop   ds
    End;
    Inc(By,320);                          { Update screen pointers }
    Dec(Ty,320);
  End;
End;

{  Main procedures  }

Procedure Update_World(X,Y,Angle: Integer);
Var Distance,CosValue: Longint;
    R_Angle: Integer;
    Height,StartY,EndY: Word;
    Shade,Counter,Find: Byte;
    Update_Floor: Boolean;
    Yoffset: Byte;
Begin
  R_Angle := (Angle-Angle_30);            { Start at Angle-30 degrees }
  If R_Angle<Angle_0 then Inc(R_Angle,Angle_360);

  For Loop1 := LeftX to RiteX do          { Draw # of collums (clipping!) }
  Begin
    Counter := 1;                         { Reset transparent wall counter }

    Repeat
      Xdist := MaxLongInt;                { Set distances to max length }
      Ydist := MaxLongInt;

      TextX := Xray(X,Y,R_Angle);         { Cast x,y rays towards walls }
      TextY := Yray(X,Y,R_Angle);         { (distance,texture numbers) }

      If Xdist<Ydist then                 { X wall is closer; store it }
      Begin
        Walls[Counter].Dist := Xdist;
        If TextX=10 then Walls[Counter].Texture := Textures[1] { Different texture on X side }
                    else Walls[Counter].Texture := Textures[TextX];
        Walls[Counter].Xoffset := RayXofs;
        Walls[Counter].Xmap := XmapX;
        Walls[Counter].Ymap := YmapX;
        Walls[Counter].MapValue := TextX;
        MapNr := TextX;
      End else
      Begin
        Walls[Counter].Dist := Ydist;
        If TextY=10 then Walls[Counter].Texture := Textures[8] { Different texture on Y side }
                    else Walls[Counter].Texture := Textures[TextY];
        Walls[Counter].Xoffset := RayYofs;
        Walls[Counter].Xmap := XmapY;
        Walls[Counter].Ymap := YmapY;
        Walls[Counter].MapValue := TextY;
        MapNr := TextY;
      End;

      If MapNr>10 then With Walls[Counter] do  { 10 or higher? Transparent wall/door }
      Begin
        If MapNr>=Door_Code+1 then        { Non-transparent door in front of the player }
        Begin
          Find := Find_Door(Xmap,Ymap);   { Check status }
          If Find <> 0 then               { Door is opening, already open, closing; so it's transparent }
          Begin
            TempMap[Xmap,Ymap] := 0;      { Set mapvalue to 0 so we can look behind the door }
            Inc(Counter);
          End else MapNr := 0;            { Door is closed; non-transparent }
        End else
        Begin                             { It's a transparent wall or door }
          TempMap[Xmap,Ymap] := 0;
          Inc(Counter);                   { Increase wall counter }
        End;
      End;

    Until MapNr < Trans_Code;             { Until wall isn't transparent (the last wall) }

    CosValue := Cosine^[(Angle_330+Loop1) mod Angle_360];
    Update_Floor := True;

    For Loop2 := Counter downto 1 do With Walls[Loop2] do
    Begin                                 { Draw all wallslices from back to front }
      Distance := (Dist*CosValue) shr 10;
      Shade := Distance shr ShadeFactor;
      Height := 16384 div Distance;
      StartY := Horizon - (Height shr 1);
      Yoffset := 0;

      If MapValue>=Door_Code then         { About to draw a door; check activity }
      Begin
        Find := Find_Door(Xmap,Ymap);     { If active, set y-offset (how much do we draw) }
        If Find = 1 then Yoffset := Doors[Door_Number].Yoffset;
      End;

      Draw_Wall_Slice(Loop1,StartY,Height,Xoffset,Shade,Yoffset,Texture,Vaddr1);
      TempMap[Xmap,Ymap] := MapValue;     { Reset original map value (for transparency) }

      If Update_Floor then                { Only draw floorslice once from farthest wall }
      Begin
        EndY := Horizon + (Height shr 1);
        Draw_FloorCeiling_Slice(CosValue,StartY,EndY,Loop1,X,Y,R_Angle,Vaddr1);
        Update_Floor := False;
      End;
    End;

    R_Angle := (R_Angle+1) mod Angle_360;
  End;
End;

{  Misc routines  }

Procedure Load_Texture(FromFile: String; ToPointer: Pointer);
Var F: File;
Begin
  Assign(F,FromFile);                     { No checking! }
  Reset(F,1);
  Blockread(F,ToPointer^,Texture_Size);
  Close(F);
End;

Procedure Load_Textures;
Var F: File;
Begin
{ E.g 1..10=walls+floors   11..15=transparent   16..20=characters}
  Load_Texture('wall01.dat',TexPoint[1]);
  Load_Texture('wall02.dat',TexPoint[2]);
  Load_Texture('wall07.dat',TexPoint[3]);
  Load_Texture('wall04.dat',TexPoint[4]);
  Load_Texture('wall06.dat',TexPoint[5]);
  Load_Texture('flr8.dat',TexPoint[6]);
  Load_Texture('wall03.dat',TexPoint[7]);
  Load_Texture('wall05.dat',TexPoint[8]);
  Load_Texture('floor4.dat',TexPoint[9]);

  Load_Texture('wall03.dat',TexPoint[Trans_Code]);
  Load_Texture('door01.dat',TexPoint[Door_Code+0]);   { 12=transparent door }
  Load_Texture('door05.dat',TexPoint[Door_Code+1]);
  Load_Texture('wall02.dat',TexPoint[Door_Code+2]);

  Assign(F,'palette.dat');
  Reset(F,1);
  Blockread(F,Palette,768);
  Close(F);
  For Loop1 := 0 to 255 do
     Set_Color(Loop1,Palette[(Loop1*3)+1] shr 2,Palette[(Loop1*3)+2] shr 2,Palette[(Loop1*3)+3] shr 2);
End;

{  Main program  }

Procedure CheckMove(X,Y: Byte);           { Lame collision detection code }
Var FindDoor: Byte;
Begin
  If TempMap[X,Y] = 0 then                { We may walk here }
  Begin
    Px := Tx;
    Py := Ty;
  End else If TempMap[X,Y] >= Door_Code then       { Found a door }
  Begin
    FindDoor := Find_Door(X,Y);
    If FindDoor <> 0 then                 { Door is active }
    Begin
      If Doors[Door_Number].Status=2 then { Door is completely open }
      Begin
        Px := Tx;
        Py := Ty;
      End;
    End;
  End;
End;

Begin
  Clrscr;                                 { Misceleanous routines }
  Randomize;
  Setup_Memory;
  Calculate_Tables;

  For Loop1 := 0 to 31 do                 { Load level into temporary 2d map }
    For Loop2 := 0 to 31 do TempMap[Loop1,Loop2] := WallMap[Loop1,Loop2];

  Videomode($13);                         { Videomode setup }

  Decode_Pcx('frame3.pcx',0,0,Vaddr2);    { Load background }
  Load_Textures;                          { Load all textures (walls+characters) }

  Px := 64*16;                            { Set various variables }
  Py := 64*14;
  Pd := Angle_90;
  Move_Value := 0;
  ShadeFactor := 6;
  Active_Doors := 1;

  Flippage(Vaddr2,Vaddr1);                { Draw first frame }
  Update_World(Px,Py,Pd);
  Flippage(Vaddr1,Vga);

  InitMouse;
  SetMouseCursor(160,100);                { Initialize mouse }

  Repeat
    If Keypressed then                    { Get keyboard information }
    Begin
      Key := Readkey;
      If Key = ' ' then
        If Active_Doors<Max_Doors then Check_Door(Px,Py,Pd);  { Search for a door in front of us }
    End;

    MouseInfo(Mx,My,Lb,Rb);               { Get mouse information }

    If Mx<>160 then                       { Alter player angle }
    Begin
      SetMouseCursor(160,100);
      Mx := 160-Mx;
      If Mx>Turning then Mx := Turning;
      If Mx<-Turning then Mx := -Turning;
      Dec(Pd,Mx);
      If Pd<Angle_0 then Inc(Pd,Angle_360);
      If Pd>Angle_360 then Dec(Pd,Angle_360);
    End;

    If Lb then If Move_Value < Speed then Inc(Move_Value);      { Forward }
    If not Lb then If Move_Value > 0 then Dec(Move_Value);
    If Rb then If Move_Value > -Speed then Dec(Move_Value);     { Backwards }
    If not Rb then If Move_Value < 0 then Inc(Move_Value);

    If Move_Value <> 0 then
    Begin
      Tx := Px + (Cosine^[Pd] * Move_Value) shr 10;  { Calculate new position }
      Ty := Py + (Sine^[Pd] * Move_Value) shr 10;
      If (Ty<Py) and (Tx<Px) then CheckMove((Ty-Close_Wall) shr 6,(Tx-Close_Wall) shr 6);
      If (Ty>Py) and (Tx>Px) then CheckMove((Ty+Close_Wall) shr 6,(Tx+Close_Wall) shr 6);
      If (Ty>Py) and (Tx<Px) then CheckMove((Ty+Close_Wall) shr 6,(Tx-Close_Wall) shr 6);
      If (Ty<Py) and (Tx>Px) then CheckMove((Ty-Close_Wall) shr 6,(Tx+Close_Wall) shr 6);
      If (Ty=Py) and (Tx>Px) then CheckMove(Ty shr 6,(Tx+Close_Wall) shr 6);
      If (Ty<Py) and (Tx=Px) then CheckMove((Ty-Close_Wall) shr 6,Tx shr 6);
      If (Ty=Py) and (Tx<Px) then CheckMove(Ty shr 6,(Tx-Close_Wall) shr 6);
      If (Ty>Py) and (Tx=Px) then CheckMove((Ty+Close_Wall) shr 6,Tx shr 6);
    End;

    Update_Doors;                         { Update data on active doors }

    Flippage(Vaddr2,Vaddr1);              { Draw the new world }
    Update_World(Px,Py,Pd);
    Flippage(Vaddr1,Vga);

  Until Key = #27;

  Release_Memory;                         { Quit program }

  Dispose(Cosine);
  Dispose(Sine);

  Videomode($3);
  Writeln('Raycasting engine by Outlaw Triad 1998'); Writeln;
  Writeln('Code - Vulture/OT');
  Writeln(' Gfx - Sappy/OT');
End.
